@@ -3,7 +3,9 @@
 
 import array
 import collections
+import functools
 import math
+import operator
 import random
 import statistics
 import typing
@@ -12,21 +14,11 @@ from . import tables, datasets
 
 try:
     import numpy
-    from numpy import array, prod, sum, take, zeros
-
-    def sumtake(a, indices):
-        return numpy.sum(numpy.take(a, indices))
-
 except ImportError:
     numpy = None
-    from ._npcompat import array, prod, take, zeros
-
-    def sumtake(a, indices):
-        return sum(map(a.__getitem__, indices))
-
 
 __all__ = ["Peptide", "tables", "datasets"]
-__version__ = "0.2.0"
+__version__ = "0.2.1"
 __author__ = "Martin Larralde <martin.larralde@embl.de>"
 __license__ = "GPLv3"
 __credits__ = """
@@ -252,9 +244,9 @@ class Peptide(typing.Sequence[str]):
         self.sequence: str = sequence
         # store an encoded version of the sequence as an array of indices
         encoder = {aa:i for i,aa in enumerate(self._CODE1)}
-        self.encoded: array = zeros(len(sequence), dtype='B')
+        self.encoded = array.array('B')
         for i, aa in enumerate(sequence):
-            self.encoded[i] = encoder.get(aa, encoder["X"])
+            self.encoded.append(encoder.get(aa, encoder["X"]))
 
     def __len__(self) -> int:
         return len(self.sequence)
@@ -317,12 +309,19 @@ class Peptide(typing.Sequence[str]):
             mu = statistics.mean(table.values())
             sigma = statistics.stdev(table.values())
             table = {k: (v - mu) / sigma for k, v in table.items()}
+        # build look up table
+        lut = [table.get(aa, 0.0) for aa in self._CODE1]
         # compute using Cruciani formula
-        s1 = s2 = 0.0
-        for aa1, aa2 in zip(self.sequence, self.sequence[lag:]):
-            s1 += table.get(aa1, 0.0) * table.get(aa2, 0.0)
-            s2 += table.get(aa1, 0.0) ** 2
-        # return correlation
+        if numpy is None:
+            s1 = s2 = 0.0
+            for aa1, aa2 in zip(self.encoded[:-lag], self.encoded[lag:]):
+                s1 += lut[aa1] * lut[aa2]
+                s2 += lut[aa1] ** 2
+        else:
+            v1 = numpy.take(lut, self.encoded[:-lag])
+            v2 = numpy.take(lut, self.encoded[lag:])
+            s1 = numpy.sum(v1*v2)
+            s2 = numpy.sum(v1**2)
         return s1 / s2
 
     def auto_covariance(
@@ -345,11 +344,17 @@ class Peptide(typing.Sequence[str]):
             sigma = statistics.stdev(table.values())
             table = {k: (v - mu) / sigma for k, v in table.items()}
         # build the lookup table
-        lut = array([table.get(aa, 0.0) for aa in self._CODE1])
+        lut = [table.get(aa, 0.0) for aa in self._CODE1]
         # compute correlation using Cruciani formula
-        v1 = take(lut, self.encoded[:-lag])
-        v2 = take(lut, self.encoded[lag:])
-        return sum(v1 * v2) / len(self.sequence)
+        if numpy is None:
+            s = 0.0
+            for aa1, aa2 in zip(self.encoded[:-lag], self.encoded[lag:]):
+                s += lut[aa1] * lut[aa2]
+        else:
+            v1 = numpy.take(lut, self.encoded[:-lag])
+            v2 = numpy.take(lut, self.encoded[lag:])
+            s = numpy.sum(v1*v2)
+        return s / len(self)
 
     def cross_covariance(
         self,
@@ -370,7 +375,7 @@ class Peptide(typing.Sequence[str]):
             0.0259803...
 
         """
-        # center the table if requested
+        # center the tables if requested
         if center:
             mu1 = statistics.mean(table1.values())
             sigma1 = statistics.stdev(table1.values())
@@ -378,12 +383,21 @@ class Peptide(typing.Sequence[str]):
             mu2 = statistics.mean(table2.values())
             sigma2 = statistics.stdev(table2.values())
             table2 = {k: (v - mu2) / sigma2 for k, v in table2.items()}
+
+        # build the lookup table
+        lut1 = [table1.get(aa, 0.0) for aa in self._CODE1]
+        lut2 = [table2.get(aa, 0.0) for aa in self._CODE1]
+
         # compute using Cruciani formula
-        s = 0.0
-        for aa1, aa2 in zip(self.sequence, self.sequence[lag:]):
-            s += table1.get(aa1, 0.0) * table2.get(aa2, 0.0)
-        # return correlation
-        return s / len(self.sequence)
+        if numpy is None:
+            s = 0.0
+            for aa1, aa2 in zip(self.encoded[:-lag], self.encoded[lag:]):
+                s += lut1[aa1] * lut2[aa2]
+        else:
+            v1 = numpy.take(lut1, self.encoded[:-lag])
+            v2 = numpy.take(lut2, self.encoded[lag:])
+            s = numpy.sum(v1*v2)
+        return s / len(self)
 
     # --- Sequence properties -----------------------------------------------
 
@@ -423,7 +437,7 @@ class Peptide(typing.Sequence[str]):
 
         """
         return {
-            aa:count/len(self.sequence)
+            aa:count/len(self)
             for aa,count in self.counts().items()
         }
 
@@ -494,8 +508,8 @@ class Peptide(typing.Sequence[str]):
               :doi:`10.1046/j.1365-2796.2003.01228.x`. :pmid:`12930229`.
 
         """
-        lut = array([tables.BOMAN["Boman"].get(aa, 0.0) for aa in self._CODE1])
-        return -sumtake(lut, self.encoded) / len(self.sequence)
+        lut = [tables.BOMAN["Boman"].get(aa, 0.0) for aa in self._CODE1]
+        return -_takesum(lut, self.encoded) / len(self)
 
     def charge(self, pH: float = 7, pKscale: str = "Lehninger") -> float:
         """Compute the theoretical net charge of a peptide sequence.
@@ -575,20 +589,35 @@ class Peptide(typing.Sequence[str]):
               ISBN:978-83-01-12044-3.
 
         """
-        sign_scale = tables.CHARGE["sign"]
-        scale = tables.PK.get(pKscale)
-        if scale is None:
+        # get chosen the pKa scale
+        scale_pKa = tables.PK.get(pKscale)
+        if scale_pKa is None:
             raise ValueError(f"Invalid pK scale: {scale!r}")
+        scale_sign = tables.CHARGE["sign"]
 
-        # nterm
-        charge = 1.0 / (1.0 + 10 ** (1.0 * (pH - scale["nTer"])))
-        # aa
-        for aa in self.sequence:
-            sign = sign_scale.get(aa, 0)
-            charge += sign / (1 + 10 ** (sign * (pH - scale.get(aa, 0))))
-        # cterm
-        charge += -1.0 / (1.0 + 10 ** (-1.0 * (pH - scale["cTer"])))
+        # build a look-up table for the pKa scale and the charge sign
+        lut_pKa = [scale_pKa.get(aa, 0.0) for aa in self._CODE1]
+        lut_sign = [scale_sign.get(aa, 0.0) for aa in self._CODE1]
 
+        # compute charge of each amino-acid, and sum
+        if numpy is None:
+            charge = 0.0
+            for aa in self.encoded:
+                pKa = lut_pKa[aa]
+                sign = lut_sign[aa]
+                charge += sign / (1.0 + 10 ** (sign * (pH - pKa)))
+        else:
+            pKa = numpy.take(lut_pKa, self.encoded)
+            sign = numpy.take(lut_sign, self.encoded)
+            charge = numpy.sum(sign / (1.0 + 10**(sign * (pH - pKa))))
+
+        # add charge for C-terminal and N-terminal ends of the peptide
+        if "nTer" in scale_pKa:
+            charge += 1.0 / (1.0 + 10 ** (pH - scale_pKa["nTer"]))
+        if "cTer" in scale_pKa:
+            charge += -1.0 / (1.0 + 10 ** (scale_pKa["cTer"] - pH))
+
+        # return the net protein charge
         return charge
 
     def hydrophobic_moment(self, window: int = 11, angle: int = 100) -> float:
@@ -628,22 +657,38 @@ class Peptide(typing.Sequence[str]):
               :doi:`10.1073/pnas.81.1.140`. :pmid:`6582470`.
 
         """
+        window = min(window, len(self))
         scale = tables.HYDROPHOBICITY["Eisenberg"]
+        lut = [scale.get(aa, 0.0) for aa in self._CODE1]
         angles = [(angle * i) % 360 for i in range(window)]
-        moment = 0.0
 
+        if numpy is None:
+            angsin = [math.sin(math.radians(theta)) for theta in angles]
+            angcos = [math.cos(math.radians(theta)) for theta in angles]
+        else:
+            angsin = numpy.sin(numpy.radians(angles))
+            angcos = numpy.cos(numpy.radians(angles))
+
+        maxnorm = 0.0
         for i in range(len(self.sequence) - window + 1):
             # compute sin and cos of angles
-            sumsin = sumcos = 0.0
-            for aa, theta in zip(self.sequence[i : i + window], angles):
-                sumsin += scale.get(aa, 0.0) * math.sin(math.radians(theta))
-                sumcos += scale.get(aa, 0.0) * math.cos(math.radians(theta))
-            # compute hydrophobic moment of window
-            hm = math.sqrt(sumsin ** 2 + sumcos ** 2) / window
-            if hm > moment:
-                moment = hm
+            if numpy is None:
+                sumsin = sumcos = 0
+                for aa, s, c in zip(self.encoded[i:i+window], angsin, angcos):
+                    sumsin += lut[aa]*s
+                    sumcos += lut[aa]*c
+            else:
+                hvec = numpy.take(lut, self.encoded[i:i+window])
+                sumsin = numpy.sum(hvec * angsin)
+                sumcos = numpy.sum(hvec * angcos)
+            # compute only the distance component (this way we can avoid
+            # computing the square root in each iteration)
+            norm = sumsin**2 + sumcos**2
+            if norm > maxnorm:
+                maxnorm = norm
 
-        return moment
+        # compute the angular moment from the norm
+        return math.sqrt(maxnorm) / window
 
     def hydrophobicity(self, scale: str = "KyteDoolittle") -> float:
         """Compute the hydrophobicity index of a protein sequence.
@@ -903,8 +948,9 @@ class Peptide(typing.Sequence[str]):
         table = tables.HYDROPHOBICITY.get(scale)
         if table is None:
             raise ValueError(f"Invalid hydrophobicity scale: {scale!r}")
-        lut = array([table.get(aa, 0.0) for aa in self._CODE1], dtype="d")
-        return sumtake(lut, self.encoded) / len(self.sequence)
+        lut = [table.get(aa, 0.0) for aa in self._CODE1]
+        # average the weight of each amino acid
+        return _takesum(lut, self.encoded) / len(self)
 
     def instability_index(self) -> float:
         """Compute the instability index of a protein sequence.
@@ -1056,8 +1102,12 @@ class Peptide(typing.Sequence[str]):
                 f"Expected str or dict, found {aa_shift.__class__.__name__}"
             )
 
-        lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
-        return sumtake(lut, self.encoded) + scale.get("nTer", 0.0) + scale.get("cTer", 0.0)
+        # build the lookup table
+        lut = [scale.get(aa, 0.0) for aa in self._CODE1]
+        # sum the mass-shift of each amino acid
+        shift = _takesum(lut, self.encoded)
+        # return the shift with C-terminal and N-terminal ends
+        return shift + scale.get("nTer", 0.0) + scale.get("cTer", 0.0)
 
     def molecular_weight(
         self,
@@ -1104,9 +1154,9 @@ class Peptide(typing.Sequence[str]):
             raise ValueError(f"Invalid average weight scale: {average!r}")
 
         # build a look-up table
-        lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+        lut = [scale.get(aa, 0.0) for aa in self._CODE1]
         # sum the weight of each amino acid
-        mass = sumtake(lut, self.encoded)
+        mass = _takesum(lut, self.encoded)
         # add weight of water molecules
         mass += scale["H2O"]
         # add mass shift for labeled proteins
@@ -1182,9 +1232,9 @@ class Peptide(typing.Sequence[str]):
         if scale not in tables.HYDROPHOBICITY:
             raise ValueError(f"Invalid hydrophobicity scale: {scale!r}")
 
-        profile = zeros(len(self.sequence) - window + 1)
-        for i in range(len(profile)):
-            profile[i] = self[i : i + window].hydrophobicity(scale=scale)
+        profile = array.array('f')
+        for i in range(len(self.sequence) - window + 1):
+            profile.append(self[i : i + window].hydrophobicity(scale=scale))
 
         return profile
 
@@ -1213,9 +1263,9 @@ class Peptide(typing.Sequence[str]):
             maximal hydrophobic moment instead of building a profile.
 
         """
-        profile = zeros(len(self.sequence) - window + 1)
-        for i in range(len(profile)):
-            profile[i] = (
+        profile = array.array("f")
+        for i in range(len(self.sequence) - window + 1):
+            profile.append(
                 self[i : i + window].hydrophobic_moment(window=window - 1, angle=angle)
             )
 
@@ -1478,7 +1528,8 @@ class Peptide(typing.Sequence[str]):
                 # if Bayes discriminant is requested, add the logarithm of
                 # the product of all non-null eigenvalues
                 if distance == "discriminant":
-                    distances[name] += math.log(prod(einvals[1:]))
+                    product = functools.reduce(operator.prod, einvals[1:])
+                    distances[name] += math.log(product)
             if not distances:
                 raise ValueError(
                     f"Cannot use {frequencies!r} frequencies with "
@@ -1529,9 +1580,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.BLOSUM)):
             # build a look-up table
             scale = tables.BLOSUM[f"BLOSUM{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return BLOSUMIndices(*out)
 
     def cruciani_properties(self) -> CrucianiProperties:
@@ -1569,9 +1620,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.CRUCIANI)):
             # build a look-up table
             scale = tables.CRUCIANI[f"PP{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return CrucianiProperties(*out)
 
     def fasgai_vectors(self) -> FasgaiVectors:
@@ -1612,9 +1663,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.FASGAI)):
             # build a look-up table
             scale = tables.FASGAI[f"F{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return FasgaiVectors(*out)
 
     def kidera_factors(self) -> KideraFactors:
@@ -1660,9 +1711,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.KIDERA)):
             # build a look-up table
             scale = tables.KIDERA[f"KF{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return KideraFactors(*out)
 
     def ms_whim_scores(self) -> MSWHIMScores:
@@ -1707,9 +1758,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.MSWHIM)):
             # build a look-up table
             scale = tables.MSWHIM[f"MSWHIM{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return MSWHIMScores(*out)
 
     def pcp_descriptors(self) -> PCPDescriptors:
@@ -1750,9 +1801,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.PCP_DESCRIPTORS)):
             # build a look-up table
             scale = tables.PCP_DESCRIPTORS[f"E{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return PCPDescriptors(*out)
 
     def physical_descriptors(self) -> PhysicalDescriptors:
@@ -1798,9 +1849,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.PHYSICAL_DESCRIPTORS)):
             # build a look-up table
             scale = tables.PHYSICAL_DESCRIPTORS[f"PD{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return PhysicalDescriptors(*out)
 
     def protfp_descriptors(self) -> ProtFPDescriptors:
@@ -1848,9 +1899,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.PROTFP)):
             # build a look-up table
             scale = tables.PROTFP[f"ProtFP{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return ProtFPDescriptors(*out)
 
     def sneath_vectors(self) -> SneathVectors:
@@ -1873,11 +1924,11 @@ class Peptide(typing.Sequence[str]):
         Example:
             >>> peptide = Peptide("QWGRRCCGWGPGRRYCVRWC")
             >>> for i, fp in enumerate(peptide.sneath_vectors()):
-            ...     print(f"SV{i+1:<3} {fp: .4f}")
-            SV1    0.1962
-            SV2    0.0466
-            SV3    0.0405
-            SV4    0.0278
+            ...     print(f"SV{i+1:<3} {fp: .5f}")
+            SV1    0.19620
+            SV2    0.04655
+            SV3    0.04050
+            SV4    0.02775
 
         References:
             - Sneath, P. H. A.
@@ -1891,9 +1942,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.SNEATH)):
             # build a look-up table
             scale = tables.SNEATH[f"SV{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return SneathVectors(*out)
 
     def st_scales(self) -> STScales:
@@ -1933,9 +1984,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.ST_SCALES)):
             # build a look-up table
             scale = tables.ST_SCALES[f"ST{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return STScales(*out)
 
     def t_scales(self) -> TScales:
@@ -1972,9 +2023,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.T_SCALES)):
             # build a look-up table
             scale = tables.T_SCALES[f"T{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return TScales(*out)
 
     def vhse_scales(self) -> VHSEScales:
@@ -2018,9 +2069,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.VHSE)):
             # build a look-up table
             scale = tables.VHSE[f"VHSE{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return VHSEScales(*out)
 
     def z_scales(self) -> ZScales:
@@ -2063,9 +2114,9 @@ class Peptide(typing.Sequence[str]):
         for i in range(len(tables.Z_SCALES)):
             # build a look-up table
             scale = tables.Z_SCALES[f"Z{i+1}"]
-            lut = array([scale.get(aa, 0.0) for aa in self._CODE1])
+            lut = [scale.get(aa, 0.0) for aa in self._CODE1]
             # average the weight of each amino acid
-            out.append(sumtake(lut, self.encoded) / len(self))
+            out.append(_takesum(lut, self.encoded) / len(self))
         return ZScales(*out)
 
     __DESCRIPTORS = {
@@ -2082,3 +2133,16 @@ class Peptide(typing.Sequence[str]):
         "VHSE": vhse_scales,
         "Z": z_scales,
     }
+
+
+
+def _takesum(table, indices):
+    """Return the sum of values from ``table`` indexed by ``indices``.
+
+    With NumPy, equivalent to ``numpy.sum(numpy.take(table, indices))``.
+
+    """
+    if numpy is None:
+        return sum(map(table.__getitem__, indices))
+    else:
+        return numpy.sum(numpy.take(table, indices))
